/**
 * Esimerkki Cache thrashing -ilmiöstä.
 *
 * Ei ole aivan sama, miten päin asettelee algoritminsa silmukat!
 *
 * Luennoilla katsottakoon, kuinka pitkä aika taulukon summan
 * laskemiseen menee kumpaisellakin eri tavalla. Ero on merkittävä, ja
 * se johtuu siitä, että toisessa aika menee prosessorin
 * välimuistirivien vaihtamiseen laskennan sijaan.
 *
 * Tämäkin esimerkki on keinotekoinen, jotta ilmiön luonne nähdään
 * "koeputkessa". Nämä operaatiothan voitaisiin oikeasti tehdä yhdellä
 * silmukalla for(int i=0;i<n*m;i++){...}, joka vastaa tässäkin olevaa
 * oikeellista tapaa, jossa välimuisti ei ruuhkaudu. [Todellisempi
 * esimerkki olisi esim. kahden matriisin kertolasku, jossa toista
 * matriisia luettaisiin riveittäin ja toista sarakkeittain. Silloin
 * voisi olla edullisinta tallentaa sarakkeittain luettava matriisi
 * alunperinkin transpoosina, koska silloin molempien matriisien
 * muistia voitaisiin lukea riveittäin.]
 *
 * Vastaava ilmiö voi esiintyä, jos esimerkiksi oliokielessä käydään
 * läpi pitkää listaa, jossa peräkkäiset viitatut oliot ovat
 * tallennettuna sikin sokin ympäri kekomuistina käytettävää fyysistä
 * muistialuetta. Alustakirjaston oma kekomuistin hallinta voi
 * esim. yrittää siirrellä peräkkäin tallennettuja olioita lähelle
 * toisiaan. Täydellinen keon hallinta onnistuu esim. C++:lla, jossa
 * on teoriassa helppo korvata olioiden varaus- ja vapautuskoodi
 * omalla. Käytännössä alustakirjastot ovat kuitenkin nykyisellään
 * niin hyviä, että niiden kanssa on vaikea kilpailla omilla
 * virityksillä!
 *
 */
#include <stdio.h>
#include <stdlib.h>

#define N 10000

/* Täyttää (m*n) -matriisin annetulla luvulla. */
void tayta(double *taulukko, int n, int m, double luku){
  for (int i=0;i<m;++i){
    for(int j=0;j<n;++j){
      taulukko[i*n+j] = luku;
    }
  }
}

/* Läpikäynti, jossa välimuisti auttaa: */
double summa(double *taulukko, int n, int m){
  double s=0.;
  for (int i=0;i<m;i++){
    for(int j=0;j<n;j++){
      s += taulukko[i*n+j];
    }
  }
  return s;
}

/* Läpikäynti, jossa välimuisti ruuhkautuu ja suorituskyky romahtaa: */
double summaB(double *taulukko, int n, int m){
  double s=0.;
  for(int j=0;j<n;j++){
    for (int i=0;i<m;i++){
      s += taulukko[i*n+j];
    }
  }
  return s;
}


int main(int argc, char** argv){
    double *t = malloc(N*N*sizeof(double)); /* dyn. varaus N*N liukuluvulle */
    if (t==NULL) exit(1);

    tayta(t,N,N,1.0);
    double s = summa(t,N,N);
    printf("Summa on %f\n",s);

    free(t); /* eksplisiittinen vapautus; hyvätapaista tehdä aina näin. */
    return 0;
}
